package edu.udo.cs.ie.cowbots.arch;

import jason.architecture.*;
import jason.asSyntax.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.locks.ReentrantLock;

/**
 * adapter to connect the cowbot architecture with
 * an environment. new perceptions are stored in a
 * first-in first-out queue. the adatper should also
 * manage the connection to an environment in case
 * of disconnects, etc.
 * 
 * @author Thomas
 *
 */
abstract public class EnvironmentAdapter { //implements EnvironmentUpdate {
	// owning agent of the adapter.
	protected AgArch	owner;
	
	// buffers for storing incoming percepts 
	protected List<Literal>		newSimPer;	// buffer to hold new simulation data
	protected List<Literal>		newEndPer;	// buffer to hold end of simulation data
	protected List<Literal> 	newEnvPer;	// buffer to hold new percepts
	protected ReentrantLock		envLock = new ReentrantLock();
	
	protected boolean hasNewPercepts;
	
	public EnvironmentAdapter(AgArch owner) {
		this.owner = owner;
		//fifoBuffer = new LinkedBlockingQueue<List<Literal>>();
		this.hasNewPercepts = false;
		newSimPer = null;
		newEnvPer = null;
	}
	
	/**
	 * 
	 */
	public abstract int getWorldTime();
	
	
	/**
	 * returns the agent's perceptions as a list of
	 * literals. might return null if no new perception
	 * are available.
	 * @return	list of perception literals
	 */
	public List<Literal> getPercepts() {		
		List<Literal> ret = null;
		if (this.newEndPer != null) {
			ret = this.newEndPer;
			newEndPer = null;
		} else if (this.newSimPer != null) {
			ret = newSimPer;
			newSimPer = null;
		} else {
			if (this.newEnvPer != null) {
				ret = this.newEnvPer;
				this.newEnvPer = null;
			}
		}
		return ret;
	}
	
	/**
	 * aquires access to environment adapter for the calling
	 * thread. the thread might be blocked until the percepts
	 * can be aquired.
	 */
	public void acquire() {
		this.envLock.lock();
	}
	
	/**
	 * releases access lock hold by the caller. has to be called
	 * whenever a caller acquired the environmental adapter for
	 * data update or retrieval.
	 */
	public void release() {
		this.envLock.unlock();
	}
	
	/**
	 * send an action to the environment
	 * @param action
	 */
	abstract void act(String action);
	 
	/**
	 * this method is called when new environmental percepts
	 * are available.
	 * 
	 * @param per	list of new perceptions
	 */
	public void newPercepts(List<Literal> per) {
		this.acquire();
		this.newEnvPer = per;
		this.release();
		owner.getArchInfraTier().wake();
	}
	
	/**
	 * this method queues information about a new simulation (a new environment
	 * the agent competes in). an agent should always setup
	 * 
	 * @param perSim data block for simulation start
	 */
	public void newSimulation(List<Literal> perSim) {
		this.acquire();
		this.newSimPer = perSim;
		this.release();
		owner.getArchInfraTier().wake();
	}
	
	/**
	 * this method is called whenever new end of simulation data
	 * are avail.
	 * 
	 * @param perEnd data block for simulation end
	 */
	public void newEndOfSimulation(List<Literal> perEnd) {
		this.acquire();
		this.newEndPer = perEnd;
		this.release();
		owner.getArchInfraTier().wake();
	}
} 
