/**
 * 
 */
package edu.udo.cs.ie.cowbots.bdi.brf;

import jason.asSemantics.Message;
import jason.asSyntax.Literal;

import java.util.*;

import edu.udo.cs.ie.cowbots.bdi.brf.operators.ExpansionOperator;
import edu.udo.cs.ie.cowbots.util.*;
import edu.udo.cs.ie.cowbots.logic.*;
import java.util.concurrent.locks.*;


/**
 * This class holds the belief base from the
 * Cowbot-BDI-model, in which the belief base consists of information
 * objects. 
 * An information object is a pair of ELP and
 * meta-information (usually time and agent source).
 * 
 * @author Thomas, Boguslaw 
 */
public class BeliefBase {
	
	// collection of information objects. the default
	// implementation simply holds a list of io lists,
	// where each list is assigned to own unique knowledge
	// operator for editing
	private List<List<InformationObject>>	ioList;
	
	// list of logical programs the agent knows (other
	// programs that are not stored in ioList
	private List<InformationObject>	programs;
	
	// list of knowledge operators with their
	// associated belief base areas
	private Map<IKnowledgeOperator, List<InformationObject>>	opDomain;

	// queue of received, but not yet processed or
	// discarded messages.
	private Queue<Message> pendingMessages;
	
	// identity table (optional element)
	IdentTable	itab;
	
	// map operator (optional element)
	MapOperator map;
	
	// annotator
	Annotator annot;
	
	// update policy
	UpdatePolicy	upol;	
	
	// knowledge operators and their order
	private Map<IKnowledgeOperator, List<InformationObject>> kopInputBuffer;
	private Map<IKnowledgeOperator, Set<ELPLiteral>> kopDeleteBuffer;
	private List<IKnowledgeOperator> kopOrder;
	
	// storage for loaded programs
	Map<String, ELP> userprograms;
	
	// storage for belief set processing programs
	List<ELP> bsetprograms;
	
	// used for the contest: agents own position
	public ELPLiteral agIson;
	
	/**
	 * default constructor
	 */
	public BeliefBase() {
		// setup message storage
		pendingMessages = new LinkedList<Message>();
		
		// setup belief base structure
		ioList = new LinkedList<List<InformationObject>>();
		
		// setup lookup table of known operators
		opDomain = new LinkedHashMap<IKnowledgeOperator, List<InformationObject>>();
		
		// setup default program storage
		programs = new LinkedList<InformationObject>();
		
		// create annotator
		this.annot = new Annotator();
		
		// create update policy
		this.upol = new UpdatePolicy();
		kopInputBuffer = new HashMap<IKnowledgeOperator, List<InformationObject>>();
		kopOrder = new LinkedList<IKnowledgeOperator>();
		kopDeleteBuffer = new HashMap<IKnowledgeOperator, Set<ELPLiteral>>();
	
		// user program storage
		userprograms = new HashMap<String,ELP>();
		
		// belief set processors
		bsetprograms = new LinkedList<ELP>();
		
		// contest extensions
		agIson = null;
	}
	
	
	/**
	 * this method returns the annotator associated with this belief base
	 * 
	 * @return instance of an annotator
	 */
	public Annotator getAnnotator() {
		return this.annot;
	}
	
	public UpdatePolicy getUpdatePolicy() {
		return this.upol;
	}
	
	/**
	 * registers a knowledge operator to the belief base. this
	 * method is responsible for setting up a domain of
	 * information objects for new (unknown) operators
	 * 
	 * @param kop knowledge operator to register
	 */
	public void registerKnowledgeOperator(IKnowledgeOperator kop) {
		if (!opDomain.containsKey(kop)) {
			List<InformationObject> iol = new LinkedList<InformationObject>();
			this.ioList.add(iol);
			this.opDomain.put(kop, iol);
			this.kopOrder.add(kop);
			this.kopInputBuffer.put(kop, new LinkedList<InformationObject>());
			this.kopDeleteBuffer.put(kop, new HashSet<ELPLiteral>());
		} else {
			// operator already working with the bb, don't do
			// anything here
		}
		
		// is operator an identity table?
		if (kop instanceof IdentTable) {
			this.itab = (IdentTable) kop;
		}
		
		if (kop instanceof MapOperator) {
			this.map = (MapOperator) kop;
		}
	}
	
	/**
	 * returns the area of information objects within
	 * the belief base associated with a knowledge operator
	 * 
	 * @param op	knowledge operator
	 * @return		list of information objects which belong to knowledge operator op
	 */
	public List<InformationObject> getOperatorIO(IKnowledgeOperator op) {
		// look up op and return list of io
		return this.opDomain.get(op);
	}

	/**
	 * returns a list with all information objects in the belief base.
	 * 
	 * @return	list of all information objects
	 */
	public List<InformationObject> getAllIO() {
		// return argument
		List<InformationObject> ret = new LinkedList<InformationObject>();
	
		// common programs
		ret.addAll( programs );
		
		// programs processed by knowledge operators
		for (List<InformationObject> iol : ioList)
			ret.addAll(iol);
		
		return ret;
	}
	
	/**
	 * this functions stores an unprocessed message
	 * in the agents belief base
	 * 
	 * @param m		message to store
	 */
	public void addMessage(Message m) {
		this.pendingMessages.add(m);
	}
	
	/**
	 * this function returns the queue of all
	 * unprocessed messages.
	 * 
	 * @return
	 */
	public Queue<Message> getMessages() {
		return this.pendingMessages;
	}
	
	/**
	 * this method deletes all information object
	 * stored in the belief base.
	 * 
	 */
	public void clear() {
		for (List<InformationObject> iol : ioList)
			iol.clear();
		programs.clear();
	}
	

	/**
	 * returns the identifier table associated with 
	 * this belief base (optional element)
	 * @return	IdentTable instance
	 */
	public IdentTable getIdentifiers() {
		return this.itab;
	}
	
	/**
	 *  returns the map operator associated with the
	 *  belief base. in any belief base, only one
	 *  map operator may exist.
	 *  
	 *  @return map operator
	 */
	public MapOperator getMapOperator() {
		return this.map;
	}
	
	
	public void addProgram(InformationObject program) {
		this.programs.add(program);
	}
	
	
	/**
	 * adds a program from a given source into the belief
	 * base. if no source is given, the source is set to
	 * the agent itself.
	 * 
	 * @param program
	 * @param source
	 * @param time
	 */
	public void addProgram(ELP program, String source, int time) {
		this.addProgram( new InformationObject(program, new Meta(time,0)));
	}
	
	
	/**
	 * adds facts from a given source to the belief base
	 * @param facts		literals to add
	 * @param source	source of literals (null for an agent itself)
	 */
	public void addFacts( Collection<ELPLiteral> facts, String source) {
		// split facts to update buffers
		for (ELPLiteral lit : facts) {
			int time = this.annot.getAndIncreaseSubtime();
			IKnowledgeOperator op = this.upol.getOp(lit.getAtom().getPredicate());
			this.kopInputBuffer.get(op).add(this.annot.createIO(lit));
		}
		// let operators process input
		for (IKnowledgeOperator op : this.kopOrder) {
			op.update( this.kopInputBuffer.get(op), this);
			this.kopInputBuffer.get(op).clear();
		}
	}
	
	
	/**
	 * delete facts from the belief base.
	 * 
	 * @param factsToDelete
	 */
	public void deleteFacts( Set<ELPLiteral> factsToDelete) {
		for (ELPLiteral l:factsToDelete) {
			IKnowledgeOperator op = this.upol.getOp(l.getAtom().getPredicate());
			this.kopDeleteBuffer.get(op).add(l);			
		}
		for (IKnowledgeOperator op : this.kopOrder ) {
			Set<ELPLiteral> s = kopDeleteBuffer.get(op);
			if (s.size() > 0) {
				op.deleteFacts(kopDeleteBuffer.get(op));
				kopDeleteBuffer.get(op).clear();
			}
		}
	}
	
	
	/**
	 * returns a uniquely-named program that is
	 * stored in the belief base of an agent.
	 * 
	 * @param name progam name
	 * @return logical program
	 */
	public ELP getProgram(String name) {
		return this.userprograms.get(name);
	}
	
	/**
	 * adds a program with a given name that is only
	 * stored, but not used by the revision.
	 * 
	 * @param name		program name
	 * @param program	logical program
	 */
	public void addUserProgram(String name, ELP program) {
		this.userprograms.put(name, program);
	}
	
	/**
	 * adds a program that operates on the belief set
	 * of an agent. those programs are concatenated
	 * at the belief state.
	 * @param program
	 */
	public void addBeliefSetProgram(ELP program) {
		this.bsetprograms.add(program);
	}
	
	/**
	 * returns a collection of programs that should be
	 * part of the belief state, but are not part of
	 * a revision process. these programs are used to
	 * perform calculations over an answer set, and
	 * the user has to take care the head alphabets of
	 * programs are consistent with ordinary programs
	 * in the beliefs of an agent.
	 * 
	 * @return
	 */
	public List<ELP> getBeliefSetPrograms() {
		return this.bsetprograms;
	}
	
	
	/**
	 * this method returns the list of all knowledge operators
	 * in order defined in the update policy.
	 * 
	 * @return list of knowledge operators
	 */
	public List<IKnowledgeOperator> getKnowledgeOperators() {
		return this.kopOrder;
	}
	
	/**
	 * resets the belief base
	 */
	public void reset() {
		for (IKnowledgeOperator op : this.kopOrder) {
			this.kopInputBuffer.get(op).clear();
			op.reset();
		}
	}
}
