/**
 * 
 */
package edu.udo.cs.ie.cowbots.bdi.brf;

import edu.udo.cs.ie.cowbots.bdi.brf.BeliefBase;

import jason.asSemantics.Event;
import jason.asSemantics.Intention;
import jason.asSemantics.Message;
import jason.asSyntax.ListTerm;
import jason.asSyntax.Literal;
import jason.asSyntax.Term;
import jason.runtime.*;

import java.io.File;
import java.util.*;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import edu.udo.cs.ie.cowbots.bdi.brf.operators.*;
import edu.udo.cs.ie.cowbots.logic.*;
import edu.udo.cs.ie.cowbots.logic.solver.*;
import edu.udo.cs.ie.cowbots.ui.*;
import edu.udo.cs.ie.cowbots.util.GridMap;

import org.w3c.dom.*;

/**
 * This class manages the components of the epistemic state in the
 * Cowbot-BDI-model, namely the belief base, the belief state and the belief
 * set.
 * 
 * It also invokes the corresponding belief operators to
 * perform belief base changes and revising an agents beliefs.
 * 
 * @author Thomas, Boguslaw, Fabian
 */
public class CowbotRevision extends CRevisionComponent implements
		IEpistemicState {
	protected BeliefBase beliefBase;
	protected BeliefSet beliefSet;

	// instance to dlv asp solver
	protected DLV dlv;

	// list of known operator classes. each operator
	// class is associated with an unique nickname
	// (the nickname makes configurations written
	// by humans easier to setup).
	protected IInferenceOperator infop;

	// debug output
	protected boolean prn = false;
	
	// map of a simulation
	protected GridMap	scenarioMap;	

	// graphical output
	// protected BDIViewFrame ui;
	protected ICowbotUI ui;

	// always perform a full revision on
	// any belief change
	protected boolean reviseFull = false;
	
	// belief set update mode
	// set to true to copy the map into the
	// belief set
	protected boolean bsOldUpdate = true;
	
	// simulation id
	protected int simulationID = -1;
		
	// rewrite state (default: remove first term as source)
	enum LiteralRewrite {
		REWRITE_NONE,	// do not rewrite literal
		REWRITE_SOURCE	// use first term as source, but do not remove term
	}
	protected Map<String, LiteralRewrite>	rewrite = new HashMap<String, LiteralRewrite>();
	
	// operation mode, simple or extended
	protected boolean		extendedRevision = false;
	
	// difference list (used to update jason bb)
	protected List<ELPLiteral>	belAdd, belDel, exploredCells;
	
	
	protected boolean directMessageAccept = true;
	/**
	 * default constructor to create a cowbot revision instance (the setup has
	 * to be done with the init method).
	 * 
	 */
	public CowbotRevision() {
		// init default modules
		this.beliefBase = new BeliefBase();
		this.beliefSet = new BeliefSet(null,null);
		this.ui = null;
		this.scenarioMap = new GridMap();
		this.scenarioMap.setDimension(1, 1);
		this.belAdd = new LinkedList<ELPLiteral>();
		this.belDel = new LinkedList<ELPLiteral>();
		this.exploredCells = new LinkedList<ELPLiteral>();				
	}

	/**
	 * revision method that is called from jason's "brf" method
	 * 
	 * @param beliefToAdd	single literal to add to an agents beliefs
	 * @param beliefToDel	single literal to remove from an agents beliefs
	 * @param i		intention that issued the brf call (always ignored)
	 * @return tupel of lists with added and deleted literals
	 */
	@Override
	public List<Literal>[] revise_belief(Literal beliefToAdd,
			Literal beliefToDel, Intention i) {

		if (beliefToAdd != null)
			addBelief(beliefToAdd);
		
		if (beliefToDel != null)
			delBelief(beliefToDel);
		
		if (this.extendedRevision)
			this.ui.updateBeliefSet(this.beliefSet.getBeliefs(), this.beliefBase.getIdentifiers(), this.scenarioMap);
		else
			this.ui.updateBeliefSet(this.beliefSet.getBeliefs(), null, null);

		return null;
	}
	
	
	/**
	 * adds a fact to an agents beliefs
	 * 
	 * @param beliefToAdd literal to add
	 */
	protected void addBelief(Literal beliefToAdd) {
		if (!beliefToAdd.isLiteral())
			return;
		
		ELPLiteral elpLit = transformJasonToELP(beliefToAdd);
		ELPLiteral delLit = null;
		if (elpLit.isStrictNegated())
			delLit = elpLit.getAtom();
		else
			delLit = new NegLiteral(elpLit.getAtom());
		
		this.beliefBase.deleteFacts( Collections.singleton(delLit));
		this.beliefBase.addFacts( Collections.singletonList(elpLit), null);
		
		// update is in belief base anywhere, decide wether to
		// revise or dirty update
		if (this.reviseFull) {
			this.reviseState();
		} else {
			// only add
			this.beliefSet.addBelief(elpLit);
			this.beliefSet.deleteBelief(delLit);
		}
		
	}
		
	
	/**
	 * removes a literal from an agents beliefs
	 
	 * @param beliefToDel literal to remove
	 */
	protected void delBelief(Literal beliefToDel) {
		
		if (!beliefToDel.isLiteral())
			return;
		
		if (!beliefToDel.isGround())
			return;
		
		// quick delete
		ELPLiteral elpLit = transformJasonToELP(beliefToDel);
		this.beliefBase.deleteFacts(Collections.singleton(elpLit));
		this.beliefSet.deleteBelief(elpLit);
	}


	@Override
	public boolean revise_percepts(List<Literal> per, List<ELPLiteral> explored) {
		boolean ret = false;

		// note: percepts are newer rewritten
		if (per.size() == 0)
			return true;
		
		// does incoming percepts starts with simulation or time?		
		Literal l0 = per.get(0);
		if (l0.getFunctor().equalsIgnoreCase("time")) {
			this.beliefBase.getAnnotator().setWorldTime( Integer.parseInt( l0.getTerm(0).toString() ));
			if (per.size() > 1)
				this.beliefBase.agIson = this.transformJasonToELP( per.get( per.size() - 1) ); 
		}
		
		/*
		if (l0.getFunctor().equalsIgnoreCase("simulation") && (per.size() >= 2)) {
			
			this.simulationID = Integer.parseInt(l0.getTerm(0).toString());
			
			Literal l1 = per.get(1);
			
			if (l1.getFunctor().equalsIgnoreCase("worldsize")) {
				scenarioMap.setDimension(
						Integer.parseInt( l1.getTerm(0).toString() ),
						Integer.parseInt( l1.getTerm(1).toString()));
			}
		}
		*/
		
		
		// transform all literals to elp literals and
		// add them to the belief base
		this.scenarioMap.resetMovers();
		this.beliefBase.map.getExploredCells().clear();
		this.beliefBase.addFacts(this.transformJasonToELP(per),null);
		this.exploredCells.addAll(this.beliefBase.map.getExploredCells());	

		// new world tick, rebuild belief state
		//if (ret)
			reviseState();
			
		// return if world changed
		return ret;
	}

	@Override
	@Deprecated
	public List<Literal> getBeliefs(ELP elp) {

		// any extra rules to execute?
		if (elp == null) {
			return Parser.toJasonLiterals(this.beliefSet.getBeliefs());
		} else {
			BeliefSet bs = computeBeliefs(elp);
			
			Set<ELPLiteral> ret = new HashSet<ELPLiteral>();
			ret.addAll(bs.getBeliefs());
			
			if (this.extendedRevision)
				copyMap(ret);
			
			return Parser.toJasonLiterals(ret);
		}
	}
	
	@Override
	public ELP getBeliefsAsELP() {
		ELP ret = new ELP();
		Iterator<ELPLiteral> iter = this.beliefSet.getBeliefs().iterator();
		while (iter.hasNext()) {
			ELPRule r = new ELPRule(1, 0);
			r.addHead(iter.next());
			ret.addRule(r);
		}
		
		// copy map to all beliefs
		if (this.extendedRevision) {
			for (int y = 0; y < this.scenarioMap.height(); y++)
				for (int x = 0; x < this.scenarioMap.width(); x++) {
					ELPLiteral cell = this.scenarioMap.getContent(x, y);
					if (cell != null) {

						ELPRule r = new ELPRule(1,0);
						r.addHead(cell);
						ret.addRule(r);
						
					}
				}
		}
		return ret;
	}
	
	/**
	 * computes a new belief set based upon the belief state and optinal extra
	 * rules.
	 * 
	 * @param extraRules
	 *            additional elp expressions
	 * @return
	 */
	protected BeliefSet computeBeliefs(ELP extraRules) {
		BeliefSet ret = null;

		List<AnswerSet> asl = null;
		// merge everything into a big elp
		ELP elp = new ELP();

		elp = elp.merge(this.getBeliefsAsELP());
		elp = elp.merge(extraRules);

		asl = dlv.computeModels(elp, 1);
		if (asl != null) {
			ret = new BeliefSet(asl.get(0).literals, this.beliefBase.getIdentifiers());
		} else {
			System.err.println("brf error: no answer set generated!");
			elp.saveAs( "elperror"+System.nanoTime()+".txt" );
			System.exit(0);
		}

		if (ret != null)
			return ret;
		else
			return new BeliefSet(null, this.beliefBase.getIdentifiers());
	}
	

	@Override
	public IEpistemicState getEpistemicState() {
		return this;
	}

	/**
	 * internal method, usually called by the update methods after the belief
	 * base was changed. this computes a new belief state and a new belief set.
	 */
	protected void reviseState() {

		// always use reviseEx
		this.reviseEx();
		
	}

	

	
	
	/**
	 * test function, prints bb contents to stdout
	 */
	public void dumpBB() {
		Collection<InformationObject> ios = this.beliefBase.getAllIO();
		Iterator<InformationObject> iter = ios.iterator();
		while(iter.hasNext()) {
			InformationObject io = iter.next();
			ELP e = io.getELP();
			ArrayList<ELPRule> rules = e.getRules();
			for (ELPRule r : rules) {
				System.out.println(r.toString());
			}
		}		
	}

	@Override
	public void queueMessage(Message m) {
		this.beliefBase.addMessage(m);
	}

	@Override
	public Queue<Message> getMessages() {
		return this.beliefBase.getMessages();
	}

	@Override
	public GridMap getMap() {
		return this.scenarioMap;
	}

	@Override
	public BeliefSet getBeliefSet() {
		return this.beliefSet;
	}
	
	/**
	 * simple revision, more or less good enough to support
	 * the massim scenario using elp or java operators
	 */
	protected void reviseSimple() {
		// get all information objects form the bb
		List<InformationObject> iol = (beliefBase.getAllIO());

		// create belief state
		ELP elp = infop.createBeliefState(iol);
		this.ui.updateBeliefState(elp);

		// revise beliefs (call dlv)
		List<AnswerSet> models = dlv.computeModels(elp, 5);
		AnswerSet model = null;
		if (models.size() > 0) {
			model = infop.selectAnswerSet(models, null);
			this.beliefSet = new BeliefSet(model.literals, this.beliefBase.getIdentifiers());
			
			// create the map from the belief set
			for (ELPLiteral lit : this.beliefSet.getBeliefs()) {
				String functor = lit.getAtom().getAtom().getPredicate();
				if (functor.equals("ison")) {
					int x = Integer.parseInt(lit.getAtom().getTerm(1));
					int y = Integer.parseInt(lit.getAtom().getTerm(2));
					this.scenarioMap.setContent(x, y, lit);
				}
				if (functor.equals("isopen")) {
					int x = Integer.parseInt(lit.getAtom().getTerm(0));
					int y = Integer.parseInt(lit.getAtom().getTerm(1));
					this.scenarioMap.setFlag(x, y, lit);
				}
			}
		}
		
		if (this.extendedRevision)
			this.ui.updateBeliefSet(this.beliefSet.getBeliefs(), this.beliefBase.getIdentifiers(), this.scenarioMap);
		else
			this.ui.updateBeliefSet(this.beliefSet.getBeliefs(), null, null);
	}
	
	
	/**
	 * extended revision. makes extensive use of the map and
	 * identtable operators. can purge elps from the belief
	 * base if the reject set of an inference operator tells
	 * to reject something.
	 * 
	 */
	protected void reviseEx() {
		
		List<InformationObject> iol = new LinkedList<InformationObject>();
					
		// fetch all information objects
		iol.addAll( this.beliefBase.getAllIO() );

		// run the super rejection
		IInferenceOperator inf = new SURI();

		// 1: build belief state
		inf.setSavePredicates( this.beliefBase.getIdentifiers().identPredicates());
		ELP bs = inf.createBeliefState(iol);
		
		// 1b: add save rules
		for (ELP bsprog : this.beliefBase.getBeliefSetPrograms())
			bs.add(bsprog);
		
		// 1c: let any operator add their save programs
		for (IKnowledgeOperator kop : this.beliefBase.getKnowledgeOperators()) {
			
			bs.add( kop.getStateProgram() );
		}
		
	 	this.ui.updateBeliefState(bs);
		
		// 2: run dlv
		SourceList sl = new SourceList();
		sl.add(bs);
		List<AnswerSet> asl = null;
		asl = this.dlv.computeModels(sl, "-N=10000", 1);
		
		// should not happen..
		// but possible if brf is called without any beliefs
		if (asl.size() <= 0) {
			System.out.println("brf error! no answer set!");
			return;
		}				
		
		// 3: get new belief set
		Set<ELPLiteral> newBS = inf.selectAnswerSet(asl, dlv).literals;
		
		// 4: prune rejected literals from bb
		Map<String,List<ELPLiteral>> rejSet = inf.getRejectSet();
		
		// merge rej Lists into one set
		Set<ELPLiteral> allRej = new HashSet<ELPLiteral>();
		for (List<ELPLiteral> rejList : rejSet.values())
			allRej.addAll(rejList);
		
		this.pruneBeliefBase(rejSet.keySet(), allRej);
		
		// 5: store new belief set
		setNewBeliefs(newBS);
	}
	
	/**
	 * applies a new belief set to the epistemic state
	 * 
	 * @param newBS	new set of beliefs
	 */
	protected void setNewBeliefs(Set<ELPLiteral> newBS) {
		
		Set<ELPLiteral> oldBS = this.beliefSet.getBeliefs();
		this.beliefSet = new BeliefSet( newBS, this.beliefBase.getIdentifiers() );
		
		// new: let all knowledge operators process the belief set
		for (IKnowledgeOperator op : this.beliefBase.getKnowledgeOperators())
			op.processBeliefSet(this.beliefSet, this.beliefBase);
		
		// 5: calculate belief set difference (change of beliefs)
		Iterator<ELPLiteral> belIter = this.beliefSet.getBeliefs().iterator();
		belAdd = new LinkedList<ELPLiteral>();
		belDel = new LinkedList<ELPLiteral>();
		
		while (belIter.hasNext()) {						
			ELPLiteral newBel = belIter.next();
			
			if (oldBS.contains(newBel)) {
				oldBS.remove(newBel);
			} else {
				belAdd.add(newBel);
			}
		}
		
		
		belIter = oldBS.iterator();
		while (belIter.hasNext()) {
			belDel.add( belIter.next() );
		}
		
		// update ui
		this.ui.updateBeliefSet(this.beliefSet.getBeliefs(), this.beliefBase.getIdentifiers(), this.scenarioMap);	
	}
	
	
	/**
	 * this method removes facts from the belief base
	 * @param predicates	functor names of predicates to remove
	 * @param facts			set of facts to remove
	 */
	protected void pruneBeliefBase(Set<String> predicates, Set<ELPLiteral> facts) {
		for (String pred : predicates) {
			IKnowledgeOperator kop = this.beliefBase.getUpdatePolicy().getOp(pred);
			kop.deleteFacts(facts);
		}
	}

	@Override
	public List<Literal>[] getChangeOfBeliefs() {
		List<Literal>[] ret = new List[2];
		
		ret[0] = transformELP2Jason( this.belAdd );
		ret[1] = transformELP2Jason( this.belDel );
		
		return ret;
	}
	
	/**
	 * auxiliary function, transforms an elp literal
	 * into a jason literal
	 * @param ll
	 * @return
	 */
	protected List<Literal> transformELP2Jason(List<ELPLiteral> ll) {
		List<Literal> ret = new LinkedList<Literal>();
		for (ELPLiteral l : ll) {

			LiteralRewrite lrw = rewrite.get(l.getAtom().getPredicate());
			
			if (lrw == null) {
				
				ret.add( Parser.toJasonWithSourceTerm(l) );
			
			} else if (lrw == LiteralRewrite.REWRITE_SOURCE) {
				
				ret.add( Parser.toJasonWithSourceRewrite(l));
				
			} else if (lrw == LiteralRewrite.REWRITE_NONE) {
				
				ret.add( Literal.parseLiteral(l.toString()));
			}
		}
		return ret;
	}
	
	/**
	 * auxiliary function, transforms jason literals
	 * into elp literals
	 * 
	 * @param beliefToAdd
	 * @return
	 */
	protected ELPLiteral transformJasonToELP(Literal beliefToAdd) {
		ELPLiteral ret = null;
		// how should the literal be transformed?
		LiteralRewrite lrw = rewrite.get(beliefToAdd.getFunctor());
		if (lrw == null) {
			ret = Parser.parseLiteralFromPlan(beliefToAdd,
					this.beliefBase.getAnnotator().getWorldTime(),"self");
		} else if (lrw == LiteralRewrite.REWRITE_SOURCE) {
			ret = Parser.parseJLiteral(beliefToAdd);
		} else if (lrw == LiteralRewrite.REWRITE_NONE) {
			ret = Parser.parseJLiteral(beliefToAdd);
		}
		
		return ret;
	}
	
	/**
	 * auxiliary function, transforms a list of
	 * jason literals into a list of elp literals
	 * 
	 * @param literals
	 * @return
	 */
	protected List<ELPLiteral> transformJasonToELP(List<Literal> literals) {
		List<ELPLiteral> ret = new LinkedList<ELPLiteral>();
		
		for (Literal l : literals)
			ret.add( this.transformJasonToELP(l));
		return ret;
	}


	@Override
	public boolean newRevisionMode() {
		return this.extendedRevision;
	}
	
	/**
	 * experimental function, copies 
	 * @param target
	 */
	protected void copyMap(Collection<ELPLiteral> target) {
		for (int y = 0; y < this.scenarioMap.height(); y++)
			for (int x = 0; x < this.scenarioMap.width(); x++) {
				ELPLiteral ison = this.scenarioMap.getContent(x, y);
				//ELPLiteral isopen = this.scenarioMap.getFlag(x, y);
				if (ison != null) {
					target.add(ison);
				}
				// isopen is added to the belief set anywhere
			}		
	}

	/**
	 * 
	 * a very simple way to handle communication: every information
	 * send by a "tell" is accepted. the information (encoded by
	 * literals) within the message is appended to the belief
	 * base.
	 * 
	 * @return list of belief additions and deletions
	 */
	@Override
	public List<Literal>[] processMessages(String agName) {
		// get message queue
		Queue<Message> msgList = this.beliefBase.getMessages();
		Iterator<Message> msgIter = msgList.iterator();
		
		// buffered message literals
		List<Literal> ll = new LinkedList<Literal>();
		List<ELPLiteral> acceptedFacts = new LinkedList<ELPLiteral>();
		
		// process message
		while (msgIter.hasNext()) {
			Message m = msgIter.next();
			if (m.getIlForce().equalsIgnoreCase("tell")) {
				// only tell ist supported
				// only respect messages with a "messagecontent" header
							
				// get message content
				Term content = null;
				if (m.getPropCont() instanceof Term)
					content = (Term) m.getPropCont();
				else
					continue;

				// get sender
				String source = m.getSender();
				if (source.equalsIgnoreCase(agName))
					source = "self";
				else
					;
				
				if ( content.isList() ) {
					ListTerm lt = (ListTerm) content;
					
					Literal head = Literal.parseLiteral( lt.get(0).toString() );
					if (head.getFunctor().equalsIgnoreCase("messagecontent")) {

						msgIter.remove();
					} else {
						// do not process message
						continue;
					}
					
					// if we do not have a simulation, ignore the message
					// bad idea at all to rely on a simulation id, but we
					// should set is anyway..
					/* do not use this for now.. 
					if (this.simulationID == -1)
						continue;
					*/
					
					Iterator<Term> li = lt.iterator();
					
					// skip header: messagecontent (floods belief base)						
					if (li.hasNext())
						li.next();
					
					while (li.hasNext()) {
						
						Term t0 =li.next();
						Literal l0 = Literal.parseLiteral(t0.toString());
						if (!l0.hasSource())
							l0.addSource( Literal.parse(source));						
						ll.add( l0 );
					}
				} else {
					// cannot be a proper message content message
					continue;
		
				}											
			}
		}
		
		
		// assign and transform literals
		acceptedFacts = this.transformJasonToELP(ll);
		this.beliefBase.addFacts(acceptedFacts, null);		
	
		// fully revise state
		// fix: dont't do it here, let it be done next cycle.
		boolean change = false;
		//if (change && this.directMessageAccept) {
		if (change) {
			this.scenarioMap.resetMovers();
			this.reviseEx();
		}					
		
		return this.getChangeOfBeliefs();
	}

	@Override
	public List<ELPLiteral> getExploredCells() {
		return this.exploredCells;
	}

	@Override
	public int getWorldTime() {
		return this.beliefBase.getAnnotator().getWorldTime();
	}

	@Override
	public void addProgram(ELP program, String source) {
		Meta m = new Meta(this.beliefBase.getAnnotator().getWorldTime(), this.beliefBase.getAnnotator().getAndIncreaseSubtime());
		this.beliefBase.addProgram( new InformationObject(program, m));
	}

	@Override
	public void init(Settings stts, DLV dlv, ICowbotUI ui) throws Exception {

		if (stts != null) {

			// look for brf configuration
			String s = stts.getUserParameter("brfconfig");
			String brfXmlFile = "brf_cfg.xml";
			String brfUserConfig = "default";

			if (s == null) {

			} else {
				s = s.replaceAll("\"", "");
				
				String sl[] = s.split("@");
				if (sl.length == 1) {
					if (sl[0].endsWith(".xml"))
						brfXmlFile = sl[0];
					else
						brfUserConfig = sl[0];
				} else if (sl.length >= 2) {
					brfUserConfig = sl[0];
					brfXmlFile = sl[1];
				} else {
					// unsupported

				}
			}

			// init components here
			// try loading default config
			Document doc = null;
			try {
				DocumentBuilder docbld = DocumentBuilderFactory.newInstance()
						.newDocumentBuilder();
				doc = docbld.parse(new File(brfXmlFile));
			} catch (Exception e) {
				throw (new Exception("brf: missing configuration file: "
						+ brfXmlFile));
			}
			NodeList nl = doc.getElementsByTagName("brf");
			this.internalInit((Element) nl.item(0), brfUserConfig, dlv, ui);

		} else {
			throw (new Exception("brf: missing mas2j user setting for brf"));
		}

	}

	
	/**
	 * initialization routine, allows the cowbot revision to be configured with
	 * user parameters.
	 * 
	 * @param brfConfig
	 *            XML Node named "brf" containing setup data
	 * @param configName
	 *            user configuration override from mas2j
	 * @param dlv
	 *            instance of dlv asp solver
	 */
	public void internalInit(Element brfConfig, String configName, DLV dlv,
			ICowbotUI ui) throws Exception {

		this.ui = ui;

		if (dlv == null)
			throw (new Exception("brf: missing dlv"));
		// dlv instance
		this.dlv = dlv;

		// known operators
		Map<String, String> kOps = new HashMap<String, String>();
		Map<String, String> iOps = new HashMap<String, String>();

		// test config
		if (brfConfig == null)
			throw (new Exception("brf: missing configuration"));

		// load brf settings
		NodeList nl = brfConfig.getElementsByTagName("knowledgeoperator");
		if (nl.getLength() == 0) {
			throw (new Exception("brf: no knowledge operator declared"));
		} else {
			for (int i = 0; i < nl.getLength(); i++) {
				Element e = (Element) nl.item(i);
				String name = e.getAttribute("name");
				String cp = e.getAttribute("class");
				kOps.put(name, cp);
			}
		}

		nl = brfConfig.getElementsByTagName("inferenceoperator");
		if (nl.getLength() == 0) {
			throw (new Exception("brf: no inference operator declared"));
		} else {
			for (int i = 0; i < nl.getLength(); i++) {
				Element e = (Element) nl.item(i);
				String name = e.getAttribute("name");
				String cp = e.getAttribute("class");
				iOps.put(name, cp);
			}
		}

		// look for configuration
		String loadConfig = (configName == null) ? "default" : configName;
		//System.out.println("cfg: "+loadConfig);
		// try using the configuration
		String update = null;
		String revision = null;

		// find config node
		nl = brfConfig.getElementsByTagName("config");
		Element cfg = null;

		for (int i = 0; i < nl.getLength(); i++) {
			Element e = (Element) nl.item(i);
			if (e.getAttribute("name").equals(loadConfig)) {
				cfg = e;
			}
		}

		// read config node
		if (cfg == null) {
			throw (new Exception("brf: configuration " + loadConfig
					+ " not declared"));
		} else {

			nl = cfg.getElementsByTagName("update");
			if (nl.getLength() > 0) {
				Element ce = (Element) nl.item(0);
				update = ce.getAttribute("policy");
			}

			nl = cfg.getElementsByTagName("revision");
			if (nl.getLength() > 0) {
				Element ce = (Element) nl.item(0);
				revision = ce.getAttribute("operator");
			}
		}

		// load inference operator
		try {
			this.infop = (IInferenceOperator) Class.forName(iOps.get(revision))
					.newInstance();
		} catch (Exception e) {
			throw (new Exception("brf: cannot instantiate inference: "
					+ revision + "@" + iOps.get(revision)));
		}
		
		// search for update policy
		nl = brfConfig.getElementsByTagName("updatepolicy");
		//System.out.println("upol: "+update);
		Element upol = null;
		for (int i = 0; i < nl.getLength(); i++) {
			Element e = (Element) nl.item(i);
			if (e.getAttribute("name").equalsIgnoreCase(update)) {
				upol = e;
				break;
			}
		}

		if (upol == null)
			throw (new Exception("brf: update policy " + update + " not found"));
		
		// read update policy
		nl = upol.getElementsByTagName("operator");
		IKnowledgeOperator perDef = null;
		IKnowledgeOperator plaDef = null;
		for (int i = 0; i < nl.getLength(); i++) {

			// read operator setup
			Element kop = (Element) nl.item(i);
			Element kconf = null;
			String name = kop.getAttribute("name");
			String trig = null;
	
			NodeList nl2 = kop.getElementsByTagName("trigger");
			if (nl2.getLength() > 0) {
				Element e = (Element) nl2.item(0);
				trig = e.getTextContent().trim();
			}
			
			NodeList nl3 = kop.getElementsByTagName("config");
			if (nl3.getLength() > 0) {
				kconf = (Element) nl3.item(0);
			}

			// setup operator
			String clname = kOps.get(name);
			if (clname == null)
				throw new Exception(
						"brf: unknown knowledge operator specified in update policy: "
								+ name);
			IKnowledgeOperator k = (IKnowledgeOperator) Class.forName(clname)
					.newInstance();
			k.init(kconf);
			this.beliefBase.registerKnowledgeOperator(k);
			
			//this.kopInputBuffer.put(k, new LinkedList<IInformationObject>());
			//this.kopOrder.add(k);
			
			if (k instanceof MapOperator) {
				MapOperator mo = (MapOperator) k;
				mo.assignMap( this.scenarioMap );
			}
		
			if (trig != null) {
				this.beliefBase.getUpdatePolicy().addOp(k, trig);
			}
			
		}
	
		perDef = new ExpansionOperator();
		this.beliefBase.registerKnowledgeOperator(perDef);
		//this.kopInputBuffer.put(perDef, new LinkedList<IInformationObject>());
		plaDef = perDef;
		//this.kopOrder.add(perDef);
		
		this.beliefBase.getUpdatePolicy().setDefaults(perDef);

		nl = upol.getElementsByTagName("norewrite");
		String nrw = "";
		for (int i = 0; i < nl.getLength(); i++) {
			Element elem = (Element) nl.item(i);
			nrw += elem.getTextContent();
			
		}
		
		String nrwl0[] = nrw.split(",");
		LinkedList<String> nrwl = new LinkedList<String>();
		for (String s0 : nrwl0) {
			String s1 = s0.trim();
			if (s1.length() > 0) {
				nrwl.add(s1);
				rewrite.put(s1, LiteralRewrite.REWRITE_NONE);
			}
		}
		
		nl = upol.getElementsByTagName("onlyannot");
		nrw = "";
		for (int i = 0; i < nl.getLength(); i++) {
			Element elem = (Element) nl.item(i);
			nrw += elem.getTextContent();
		}
		nrwl0 = nrw.split(",");
		nrwl = new LinkedList<String>();
		for (String s0 : nrwl0) {
			String s1 = s0.trim();
			if (s1.length() > 0) {
				nrwl.add(s1);
				rewrite.put(s1, LiteralRewrite.REWRITE_SOURCE);
			}
		}
		
		// if a map and identtable operator is present, the revision
		// may enter extended mode
		if ((this.beliefBase.getIdentifiers() != null)
			&& (this.beliefBase.getMapOperator() != null))
			this.extendedRevision = true;
		
		//System.out.println("no rewrite:" +nrwl);
		this.scenarioMap.setIdentTable(this.beliefBase.getIdentifiers());
	}

	@Override
	public ELP getProgram(String name) {
		return this.beliefBase.getProgram(name);
	}

	@Override
	public void addUserProgram(String name, ELP program) {
		this.beliefBase.addUserProgram(name, program);
	}

	@Override
	public BeliefBase getBeliefBase() {
		return this.beliefBase;
	}

	@Override
	public void reset() {
		this.scenarioMap.resetMovers();
		this.beliefBase.reset();
		this.beliefSet = new BeliefSet(null,null);
	}
	
}
