package edu.udo.cs.ie.cowbots.bdi.desgen;

import jason.asSyntax.Literal;
import jason.runtime.Settings;
import jason.stdlib.desire;
import jason.stdlib.string;

import java.util.List;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.FileReader;
import java.io.IOException;
import java.util.Collections;

import edu.udo.cs.ie.cowbots.bdi.brf.IEpistemicState;
import edu.udo.cs.ie.cowbots.logic.solver.DLV;

/**
 * Controller class for generating new desires. Uses a static
 * set of motives to update the agents desirestate. 
 * 
 * @author dhoelzgen
 *
 */
public class DesireComponent {

	protected IMotiveSet motiveSet;
	protected IDesireState desireState;

	/**
	 * Empty constructor for DesireComponent.
	 */
	public DesireComponent() {
		this(new MotiveSet());
	}

	/**
	 * Constructor for DesireComponent which allows to specify
	 * the agents set of motives.
	 * 
	 * @param motiveSet
	 */
	public DesireComponent(IMotiveSet motiveSet) {
		this.motiveSet = motiveSet;
		this.desireState = new DesireState();
	}
	
	/**
	 * Constructor for DesireComponent, which loads the
	 * agents set of motives from a path provided in 
	 * the settings. 
	 * 
	 * @param stts Settings from mas2j file
	 */
	public DesireComponent(Settings stts) {
		this();
		
		if (stts != null) {
			String motivePath = stts.getUserParameter("motives");
			motivePath = motivePath.replace("\"", "");
			if (motivePath != null) {
				File motiveDir = new File(motivePath);
				if (!motiveDir.isDirectory())
					System.err.println("Motive Dir ("+ motivePath + ") is not a directory. Please check parameters in mas2j file.");
				else
					this.motiveSet = this.loadMotivesFromDirectory(motiveDir);
			} else {
				System.err.println("Could not load motive set. Please check parameters in the mas2j file.");
			}

		}		
	}

	/**
	 * Provides access to the agents current desirestate.
	 * 
	 * @return Current agents desirestate
	 */
	public IDesireState getDesireState() {
		return desireState;
	}

	/**
	 * Updates the agents desirestate based on provided
	 * beliefs.
	 * 
	 * @param Current agents beliefs to update desirestate
	 * @return The updated desirestate
	 */
	public IDesireState updateDesireState(IEpistemicState bel) {

		// Generate new desires, depending on provided agents beliefs
		// As beeing a set, the desireState should not overide existing
		// motives
		for (IMotive motive : this.motiveSet) {
			List<CDesire> newDesires = motive.generateDesire(bel);
			for (CDesire des : newDesires) {
				if (!this.desireState.contains(des))
					this.desireState.add(des);
			}
		}

		try {
			
			// An intensity equal or less zero causes the deletion of
			// the desire. Because we cannot change the state during
			// the update process, we need a temporary one.
			IDesireState tmpDesireState = this.desireState.getClass().newInstance();

			// Intensity update for all current desires, followed by
			// proof for mitigation
			for (CDesire desire : this.desireState) {
				desire.updateIntensity(bel);
				desire.updateMitigation(bel);

				// If the intensity is still more than zero, keep it in
				// the temporary desirestate
				if (desire.getIntensity() > 0) tmpDesireState.add(desire);
			}
			
			// Replace desirestate with temporary one
			this.desireState = tmpDesireState;
			
		} catch (Exception e) {
			// Error while creating temporary desire state. Will not
			// update and return old desirestate...
			// TODO: Error Handling
			System.err.println("Could not update desire state: " + e.toString());
		}

		return getDesireState();
	}

	/**
	 * Enum used for reading the motive files.
	 * 
	 * @author dhoelzgen
	 *
	 */
	protected enum ReadState {
		waiting,
		generate,
		update,
		mitigate,
		uniquename
	}
	
	/**
	 * Reades all *.motive files in the given directory to 
	 * initialize the agents motives
	 * @param dir
	 */
	protected IMotiveSet loadMotivesFromDirectory(File dir) {
		MotiveSet rMotiveSet = new MotiveSet();
		
		// Get all files that end with .motive
		File[] motiveFiles = dir.listFiles(
				new FileFilter()
				{
					public boolean accept(File pathname) {
						if (pathname.getName().endsWith(".motive")) return true;
						else return false;
					}
				});
		
		for (File motiveFile : motiveFiles) {
			StringBuilder generateELP = new StringBuilder();
			StringBuilder updateELP = new StringBuilder();
			StringBuilder mitigateELP = new StringBuilder();
			StringBuilder uniqueName = new StringBuilder();
			
			// Read the three config file sections and put them to the
			// matching string builders
			try {
				BufferedReader in = new BufferedReader(new FileReader(motiveFile.getAbsolutePath()));
				String readLine = null;
				ReadState rState = ReadState.waiting;
				while ((readLine = in.readLine()) != null) {
					if (readLine.startsWith("#")) {
						if (readLine.toLowerCase().endsWith(" end")) rState = ReadState.waiting;
						else if (readLine.toLowerCase().startsWith("#desgen")) rState = ReadState.generate;
						else if (readLine.toLowerCase().startsWith("#intupdate")) rState = ReadState.update;
						else if (readLine.toLowerCase().startsWith("#intmitigation")) rState = ReadState.mitigate;
						else if (readLine.toLowerCase().startsWith("#uniquename")) rState = ReadState.uniquename;
						else System.err.println("Unrecognized keyword in motive file " + motiveFile.getName()
								+ ": " + readLine);
					} else {
						switch (rState) {
							case generate:
								generateELP.append(readLine); break;
							case update:
								updateELP.append(readLine); break;
							case mitigate:
								mitigateELP.append(readLine); break;
							case uniquename:
								uniqueName.append(readLine); break;
							default: break;
						}
					}
				}
			} catch (IOException e) {
				e.printStackTrace();
			}
			
			// Generate the new motive and add it to the returned motive set
			rMotiveSet.add(new DefaultMotive(motiveFile.getName(),
					generateELP.toString(), updateELP.toString(), mitigateELP.toString(), uniqueName.toString()));
			
		}
		
		return rMotiveSet;
	}
}
