package edu.udo.cs.ie.cowbots.elpParser.dlp;


import jason.stdlib.string;

import java.util.*;

import edu.udo.cs.ie.cowbots.logic.ELPRule;

/**
 * this class extends from its father class ELPRule.
 * stand by parsing
 * 
 * @author Wei
 * @see ELPRule
 */
public class DLPRule extends ELPRule{
	
	ArrayList<DLPLiteral>	head;
	ArrayList<DLPLiteral>	body;
	
	/**
	 * returns true whether the rule has only one or more literal for head and
	 * they are facts, but no body literals.
	 * @return {@link Boolean} {@link DLPLiteral}
	 */
	public boolean isFact() {
		
		if(isDisjunctive())
			for (DLPLiteral aHead : head) {
				aHead.LitIsAfact(true);
			}
		
		if (head.size() > 0)
			head.get(0).LitIsAfact(true);
		
		return body.size() == 0;
	}
	
	/**
	 * returns true if the rule is a constraint (no head literals)
	 * and at least one literal(it is indifferently true or negative or strict negative)
	 * @return {@link Boolean}
	 */
	public boolean isConstraint() {
		return head.size() == 0 && body.size()>=1;
	}
	
	/**
	 * returns true if it is a complete rule what means
	 * at least one literal for body and one for head
	 * @return {@link Boolean}
	 */
	public boolean isRule() {
		return head.size()>=1 && body.size()>=1;
	}
	
	/**
	 * indicated whether there is one more literals
	 * in the read of the rule.
	 * @return {@link Boolean}
	 */
	public boolean isDisjunctive() {
		if (head != null)
			return head.size() > 1;
		else
			return false;
	}
	
	/**
	 * indicated whether there is one more literals
	 * in the read of the rule.
	 * @return {@link Boolean}	 
	 */
	public boolean isConjunctive() {
		if (body != null)
			return body.size() > 1;
		else
			return false;
	}
	
	/**
	 * Create a default rule.
	 * @see DLPRule
	 */
	public DLPRule() {
		this.head = new ArrayList<DLPLiteral>(1);
		this.body = new ArrayList<DLPLiteral>(1);
	}
	
	
	/**
	 * An user defined rule with prospective
	 * body and head length.
	 * @param maxHead {@link Integer}
	 * @param maxBody {@link Integer}
	 * @see DLPRule
	 */
	public DLPRule(int maxHead, int maxBody) {
		this.head = new ArrayList<DLPLiteral>(maxHead);
		this.body = new ArrayList<DLPLiteral>(maxBody);
	}
	
	/**
	 * creates new rule which prepare to change its content, 
	 * based on a existed rule.	 
	 * 
	 * @param other	rule to copy {@link DLPRule}
	 * 
	 */

	public DLPRule( DLPRule other ) {
		this.head = new ArrayList<DLPLiteral>(other.nHead());
		for (int i = 0; i < other.nHead(); i++) {
			this.head.set(i, other.head.get(i));
		}
		this.body = new ArrayList<DLPLiteral>(other.nBody());
		for (int i = 0; i < other.nBody(); i++) {
			this.body.set(i, other.body.get(i));
		}
	}
	
	
	/**
	 * add a literal to head
	 * @param l {@link DLPLiteral}		
	 */
	public void addHead(DLPLiteral l) {
		this.head.add(l);
	}
	
	/**
	 * add a literal to body
	 * @param l {@link DLPLiteral}
	 */
	public void addBody(DLPLiteral l) {
		body.add(l);
	}
	
	
	public ArrayList<DLPLiteral> getBodies() {
		return body;	
	}
	
	
	public void addBodies(Collection<DLPLiteral> lits) {
		this.body.addAll(lits);
	}
	
	/**
	 * overview during parse. 
	 * */
	public void resolve(){
		
		System.out.println("Resolving one line...");
		
		if(this.isFact() && !this.isConstraint()){
			System.out.println("This is a fact.");
			resolvedFact();		}
		
		if(this.isConstraint()){
			System.out.print("This is a constraint. ");
			resolvedConstraint();
		}
		
		if(this.isRule()){
			System.out.print("This is a rule. ");
			resolvedRule();
		}
		
	}
	
	/**
	 *@see {@link DLPRule#resolve()} 
	 * */
	public void resolvedFact(){		
		System.out.println("Fact: "+head.get(0).toString());		
	}
	
	/**
	 *@see {@link DLPRule#resolve()}  
	 * */
	public void resolvedConstraint(){
				
		if(this.isConjunctive()){
			System.out.println("And its body is conjunctive.");
			int b=0;
			for (DLPLiteral aBody : body) {
		    System.out.println("Body"+b+" : "+aBody.toString());
		    b++;
			}			
		}else{
			System.out.println();
			System.out.println("Body: "+body.get(0).toString());
		}	
	}
	
	/**
	 *@see {@link DLPRule#resolve()}
	 * */
	public void resolvedRule(){
		
		if(this.isDisjunctive()){
			System.out.println("And its head is disjunctive.");
			int h=0;
			for (DLPLiteral aHead : head) {
		    System.out.println("head"+h+" : "+aHead.toString());
		    h++;
			}			
		}else{
			System.out.println();
			System.out.println("head: "+head.get(0).toString());
		}		
		resolvedConstraint();
	}
	
	/**
	 *@see {@link ELPRule#toString()}
	 */
	protected String toDlpString() {

		return super.toString();
		
	}
	
	/**
	 * internal helper function to dump a list of literals
	 * @param ll	list of literals 
	 * @param sep	literal separator {@link string}
	 * @return		present one line rule {@link string}
	 */
	protected String dumpLiteralList(ArrayList<DLPLiteral> ll, String sep) {
		
		if (ll == null)
			return null;
		if (ll.size() < 1)
			return null;
		String ret = ll.get(0).toString();
		for (int i = 1; i < ll.size(); i++) {
			ret +=sep+ll.get(i).toString();
		}
		return ret;
		
	}
	
	/**
	 * returns the size of head
	 * @return	number of head literals
	 */
	public int nHead() {
		return head.size();
	}
	
	/**
	 * returns the size of body, including arithmetical expressions.
	 * @return	number of body literals {@link Integer}
	 */
	public int nBody() {
		return body.size();
	}
	
	/**
	 * returns the i-literal in head
	 * @param  literal index {@link Integer}
	 * @return	i-literal in head {@link DLPLiteral}
	 */
	public DLPLiteral getHead(int i) {
		return head.get(i);
	}
	
	/**
	 * returns the i-literal in Body
	 * @param literal index {@link Integer}
	 * @return	i-literal in body {@link DLPLiteral}
	 */
	public DLPLiteral getBody(int i) {
		return body.get(i);
	}
}
