package edu.udo.cs.ie.cowbots.launcher;

import java.io.File;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.NodeList;

import java.util.*;

/**
 * utility class that parses a local configuration file
 * for setting up a mas2j. used during initialization by
 * various classes (environment, agent, agentarch).
 * 
 * @author Thomas
 *
 */
public class CowbotsConfig {

	Element				config = null;
	
	String				dlvPath = null;
	String				clingoPath = null;
	String				massimPath = null;
	String				host = null;
	List<String[]>		allhosts = new LinkedList<String[]>();
	
	// list of deliberation options
	HashMap<String,String>	delibOptions = new HashMap<String,String>();
	HashMap<String,String>	agOptions = new HashMap<String,String>();
	HashMap<String,HashMap<String,String>> allOptions = new HashMap<String,HashMap<String,String>>();
	HashMap<String,String>	desgenOptions = new HashMap<String,String>();
	
	static		CowbotsConfig 	instance = null;
	/**
	 * initializes the local configuration from
	 * the local_config.xml file
	 */
	protected CowbotsConfig() {
		allOptions.put("deliberation", delibOptions);
		allOptions.put("agclass",agOptions);
		allOptions.put("desgen",desgenOptions);
		
		DocumentBuilder docbld;
		try {
			docbld = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			Document doc = docbld.parse(new File("local_config.xml"));			
			NodeList nl = doc.getElementsByTagName("localconfig");
			if (nl.getLength() > 0) {
				config = (Element) nl.item(0);
			}
			if (config != null) {
				nl = config.getElementsByTagName("config");
				for (int i = 0; i < nl.getLength(); i++) {
					Element ce = (Element) nl.item(i);
					String name = prepString(ce.getAttribute("name"));
					if (name != null) {
						// setup dlv
						if (name.equalsIgnoreCase("dlv")) {
							this.dlvPath = prepString(ce.getAttribute("path"));
						}
						
						// setup clingo
						if (name.equalsIgnoreCase("clingo")) {
							this.clingoPath = prepString(ce.getAttribute("path"));
						}
						
						// local massim environment
						if (name.equalsIgnoreCase("massim")) {
							this.massimPath = prepString(ce.getAttribute("path"));
						}
						
						// delib options
						if (name.equalsIgnoreCase("deliberation")) {

							String option = prepString(ce.getAttribute("norestart"));
							if (option != null) {
								this.delibOptions.put("norestart", option);
								System.out.println("[agclass] delib norestart: "+option);
							}
						}
						
						// cowbot agent class options
						if (name.equalsIgnoreCase("agclass")) {
							String option = prepString(ce.getAttribute("msgproc"));
							if (option != null) {
								this.agOptions.put("msgproc",option);
								System.out.println("[agclass] message handler: "+option);
							}
							option = prepString(ce.getAttribute("printinfo"));
							if (option != null) {
								this.agOptions.put("printinfo", option);
								System.out.println("[agclass] print info: "+option);
							}
						}
						
						// desire generation options
						if (name.equalsIgnoreCase("desgen")) {
							// default cluster algorithm override
							String option = prepString(ce.getAttribute("clusteralgorithm"));
							if (option != null) {
								this.desgenOptions.put("clusteralgorithm", option);
								System.out.println("[desgen] default clusterer: "+option);
							}
						}
					}
									
					String host = prepString(ce.getAttribute("host"));
					if (host != null) {
						String show = prepString(ce.getAttribute("button"));
						String s2[] = new String[2];
						s2[0] = host;
						s2[1] = show;
						this.allhosts.add(s2);
					}
				}
			}
		} catch (Exception e) {
			System.err.println("LocalConfig initialization failure!");
			e.printStackTrace();
		}		
		
	}

	/**
	 * this method returns the global instance of the
	 * cowbots configuration
	 * 
	 * @return	CowbotsConfig instance
	 */
	public static synchronized CowbotsConfig getInstance() {
		if (instance == null)
			instance = new CowbotsConfig();
		
		return instance;
	}
	
	
	/**
	 * this methods returns the path to a dlv asp
	 * solver binary or null
	 * 
	 * @return path to dlv
	 */
	public String getDLVPath() {
		return this.dlvPath;
	}
	
	/**
	 * this method returns the path to a clingo
	 * asp solver binary or null
	 * @return path to dlv
	 */
	public String getClingoPath() {
		return this.clingoPath;
	}
	
	/**
	 * this method returns the path to a local
	 * massim svg output directory
	 * 
	 * @return	local massim svg output directory
	 */
	public String getMassimPath() {
		return this.massimPath;
	}
	
	/**
	 * this method sets the massim host for all
	 * agents.
	 * 
	 * @param host	hostname for massim environment
	 */
	public void setHost(String host) {
		this.host = host;
	}
	
	/**
	 * this method returns the host for the
	 * massim server the agents should connect
	 * to.
	 * 
	 * @return	host of massim server
	 */
	public String getHost() {
		return this.host;
	}
	
	/**
	 * this methods returns all available hosts
	 * of massim servers defined in the local_config.xml
	 * file. used by the cowbots launcher during
	 * mas startup.
	 * 
	 * @return	list of available massim servers
	 */
	public List<String[]> getHosts() {
		return this.allhosts;
	}
	
	/**
	 * auxiliary method that removes double
	 * quotation marks from a string
	 * 
	 * @param s		input string to process
	 * @return		processed string or null (if string remains empty)
	 */
	protected String prepString(String s) {
		String ret = s.replaceAll("\"", "");
		ret.trim();
		if (ret.length() < 1)
			return null;
		else
			return ret;
		
	}
	
	
	/**
	 * returns the value of an attribute for one of the freely configurable
	 * options in the localconfig class.
	 * 
	 * @param component		target component to load options for
	 * @param attribute		attribute value to query
	 * @return
	 */
	public String getOption(String component, String attribute) {
		HashMap<String,String> map0 = allOptions.get(component);
		if (map0 == null)
			return null;
		return map0.get(attribute);
	}


}
