package edu.udo.cs.ie.cowbots.logic.solver;

import jason.asSemantics.TransitionSystem;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.Iterator;
import java.util.HashSet;
import java.util.TreeSet;
import java.util.concurrent.atomic.*;
import java.util.concurrent.Semaphore;
import java.io.*;


import edu.udo.cs.ie.cowbots.logic.*;

/**
 * the dlv wrapper class, based on the one found in KiMAS - Knowledge in
 * Multi Agent Systems. 
 * 
 * @author Thomas, Elly
 *
 */
public class DLV {

	/**
	 * utility class, used by the parser
	 * to store answer set weights 
	 * @author Thomas
	 *
	 */
	class ASWeight {
		public ASWeight(int w, int l) {
			weight = w;
			level = l;
		}
		public final int weight;
		public final int level;
	}
	
	// path to dlv binary, binary inclusive
	protected String	path2dlv;
	
	// path to log error messages
	protected String	logpath;
	
	// remember operating system
	protected boolean	osLinux;
	
	// counter to count number of solver calls
	static AtomicInteger	callCounter = new AtomicInteger(0);
	
	// number of processors
	protected int		cpuCount;

	// print timing
	protected boolean	printTimings;
	
	// use semaphore
	protected boolean	useSemaphore;

	// semaphore to limit dlv instances to number of processors
	static protected 	Semaphore	sem;
	
	public DLV(String path) {
		path2dlv = path;
		logpath = "";
		osLinux = false;
		
		cpuCount = Runtime.getRuntime().availableProcessors();
		
		int cc = callCounter.getAndIncrement();
		if (cc == 0)
			sem = new Semaphore(cpuCount);
		
		//System.out.println("dlv: using "+cpuCount+" cpu(s)");
		
		// on linux, use taskset to choose cpu
		if (System.getProperty("os.name").startsWith("Linux") ) {
			osLinux = true;									
		}				
		
		printTimings=false;
		useSemaphore=false;
	}
	
	public void setLogPath(String path) {
		this.logpath = path;
	}
		
	public List<AnswerSet> computeModels(ELP program, int maxmodels) {
		
		SourceList sl = new SourceList();
		sl.add(program);
		return computeModels(sl,"",maxmodels );
	}
	
	public List<AnswerSet> computeModels(SourceList input, String options, int models) {
		ArrayList<AnswerSet> ret = new ArrayList<AnswerSet>();
		List<String> dlvout = null;
		
		// build command line
		String cmdln = path2dlv;		
		
		// add options
		cmdln += " "+options;
		
		// create dlv process
		dlvout = runDLVEx(cmdln, models, input);
		
		// check output and create answer sets
		Iterator<String> i = dlvout.iterator();
		while (i.hasNext()) {
			String s1 = i.next();
						
			// check if output is relevant 
			if (s1.startsWith("{") ) {
				// standard answer set
				ret.add(new AnswerSet( this.parseModel(s1), 0,0 ));
			} else if (s1.startsWith("Best model: {")) {
				// answer sets with weights
				this.parseModel(s1);
				ASWeight asw = null;
				
				// parse weight
				if (i.hasNext()) {	// should never fail..
					String s2 = i.next();
					asw = parseWeight(s2);
				}
			}
		}
		return ret;
	}
		
	/**
	 * auxiliary function that returns an empty string
	 * or a command for a task that runs dlv. on linux,
	 * taskset is choosen by default.
	 *  
	 * @return
	 */
	protected String getRunCmdPrefix() {
		String ret = "";
		
		if (this.osLinux) {
			int callNr = callCounter.addAndGet(1);
			int cpu = (callNr % cpuCount);
			cpu = 1;
			ret = "taskset -c "+cpu+" ";						
		}
		
		return "";
	}
	
	/**
	 * main method used for running a answer set solver (external
	 * utility program) and parsing its output.
	 * 
	 * this is a modified version of the computeModels code from
	 * the dlv class found in KiMAS - Knowledge in Multi Agent Systems, 
	 * credits to I. Drobiazko for writing original code.
	 * 
	 * @param cmdln		command to run dlv
	 * @param maxAS		maximum number of answer sets
	 * @param prog		program to pass via stdin (-- option)
	 */
	protected List<String> runDLVEx(String cmdln, int maxAS, SourceList programs) {
		String text = "";
		StringBuffer errors = new StringBuffer();
		List<String> ret = new LinkedList<String>();
		
		if (maxAS > 0)
			cmdln += " -n="+maxAS;
		
		String files = "";
		int nFileSources = 0;
		// process ELP sources
		for (ELPSource s : programs) {
			if (s.getType() == ELPSource.SRC_FILE) {
				files += " "+s.getSrcFile();
				++nFileSources;
			}
		}
		
		if (nFileSources < programs.size())
			cmdln += " --";
		cmdln += files;
		
		boolean pipeInput = nFileSources < programs.size();
		
		// apply run command prefix
		cmdln = getRunCmdPrefix() + cmdln;
		
		// measure dlv time
		long time = System.currentTimeMillis();
		// time required to aquire cpu
		long waitTime = time;
		
		// block if cpu not avail
		if (useSemaphore) {
			try {
				sem.acquire();
			} catch (Exception e) {
				System.out.println("dlv: failed to aquire semaphore");
			}
		}
		waitTime = System.currentTimeMillis()-waitTime;
		
		try {
			// setup call to dlv
			Process process = Runtime.getRuntime().exec(cmdln);
			
			BufferedReader inputStream = new BufferedReader(
					new InputStreamReader(process.getInputStream()));
			BufferedReader errorStream = new BufferedReader(
					new InputStreamReader(process.getErrorStream()));
			
			
			boolean useThread = false;
			
			BufferedWriter outputStream = null;
			if (pipeInput) {
				outputStream = new BufferedWriter(new OutputStreamWriter(process.getOutputStream()));
				for (ELPSource elps : programs) {
					if (elps.getType() != ELPSource.SRC_FILE)
						elps.addResource(outputStream);
				}
				outputStream.flush();
				outputStream.close();
				
			}
			
			if (!useThread) {
				// store result
				while ((text = inputStream.readLine()) != null) {
					ret.add(text);
				}
				
				// check for errors
				text = "";
				while ((text = errorStream.readLine()) != null) {
					errors.append(text);
				}
			} 
			
			inputStream.close();
			errorStream.close();
			
			process = null;
		} catch (IOException e) {
			System.err.println("dlv error!");
			System.err.println(text);
			System.err.println(e);
			String errFile=this.logpath + "dlv_ioerr_" + System.nanoTime() + ".txt";
			try {
			programs.saveAs(errFile);
			} catch (Exception e2) {
				
			}
			
		}				
		
		time = System.currentTimeMillis() - time;
		
		if(!"".equals(errors.toString())){
			
			String errMsg = "dlv error!\n";
			errMsg += errors + "\n";
			
			long errtime = System.nanoTime();
			String errFile = logpath + "dlverror_elp_"+errtime+".txt";
			
			errMsg += "error program saved as: " + errFile;
			System.out.println(errMsg);
			
			try {
				programs.saveAs(errFile);
				
				BufferedWriter bw = new BufferedWriter( new FileWriter( logpath+"dlverror_msg_"+errtime+".txt" ));
				bw.write(errMsg);
				bw.flush();
				bw.close();
				
			} catch (Exception e) {
				System.out.println("dlv error: could not save error log!\n"+
					e.toString());
			}
		}
		
		if (useSemaphore) {
			sem.release();
			if (printTimings) {
				long real = time - waitTime;
				System.out.println("dlv call: "+time+"ms, real:"+real+"ms, wait"+waitTime+"ms");
			}
		} else {
			if (printTimings)
				System.out.println("dlv call: " +time+"ms");
		}
		
		return ret;
	}
	
	/**
	 * simple parser function, extracts all literals from
	 * an answer set provided by dlv. if the empty answer
	 * set is found, an empty Set<ELPLiteral> is returned.
	 *  
	 * @param s		answer set from dlv
	 * @return		set of literals
	 * 
	 * @author Thomas
	 */
	protected Set<ELPLiteral> parseModel(String s) {		
		// parse result
		Set<ELPLiteral> ret = new LinkedHashSet<ELPLiteral>();
		// buffer to hold symbols during literal parsing.
		// can hold atoms up to arity 99, first token
		// is always the predicate name.
		String tokens[] = new String[100];
		int	tokencount = 0;
		int tokenStart = 0;
		int tokenEnd = 0;
		
		// current position in string
		int cursor = 0;
		boolean done = false;
		
		// literal attributes
		boolean isneg = false;
		char c = ' ', lastc = ' ';

		// scan output
		while (!done) {

			// scan input
			lastc = c;
			c = s.charAt(cursor++);
			
			// skip open par
			if (c == '{') {
				tokenStart = cursor;
				continue;
			}					
			
			// is literal negated?
			if (c == '-') {
				isneg = true;
				tokenStart = cursor;
				continue;
			} 
			
			// comma separates literals
			if ((c == ',')) {
				
				// not parsing a term list, tokenStart is valid				
				tokenEnd = cursor-1;
				if (lastc==')')
					--tokenEnd;
				tokens[tokencount++] = s.substring(tokenStart, tokenEnd).toString();
										
				ELPAtom a = new ELPAtom(tokens, tokencount);
				if (isneg)
					ret.add(new NegLiteral(a));
				else
					ret.add(a);
				
				// cursor is always at whitespace at
				// this point, compensate that
				++cursor;	// cursor now at next literal
				tokenStart = cursor;
				isneg = false;
				
				tokencount = 0;					

				continue;
			}
			
			// found opening parenthesis of a term list
			if (c == '(') {
				
				tokenEnd = cursor-1;
				tokens[tokencount++] = s.substring(tokenStart, tokenEnd);
				
				tokenStart = cursor;
				// scan term list
				while (c != ')') {
					c = s.charAt(cursor++);
					if (c == ',') {
						tokenEnd = cursor-1;
						tokens[tokencount++] = s.substring(tokenStart, tokenEnd);						
						tokenStart = cursor;						
					}
				}
				
				continue;
			}
			
			
			// closing bracelet of answer set
			if (c == '}') {
				done = true;
				tokenEnd = cursor-1;
				
				if (lastc==')')
					--tokenEnd;				
				
				// catch last literal, and handle 
				// empty answer set correctly
				if (tokenEnd > tokenStart) {
					tokens[tokencount++] = s.substring(tokenStart, tokenEnd);
					
					ELPAtom a = new ELPAtom(tokens, tokencount);
					if (isneg)
						ret.add(new NegLiteral(a));
					else
						ret.add(a);
				}
				
				tokencount = 0;
			}									
		}
		
		return ret;
	}	

	
	protected ASWeight parseWeight(String s) {
		return new ASWeight(0,0);
	}
	
	/**
	 * method used to compute plans from k planning language
	 * @param ts TODO
	 * @param savedAs TODO
	 * 
	 */
	
	public Set<ELPLiteral> computePlans(int maxint, TransitionSystem ts, String savedAs) {
		
		List<String> dlvout = null;
		Set<ELPLiteral> planset = null;
		String plan = null;
		
		//System.out.println(program.getKnowledge());

		String cmdln = path2dlv + " -FPsec -N="+ maxint +" -n=1 ";
		 //add option

		String path2k ="";
		String myself= ts.getUserAgArch().getAgName();
		
		// build command line TODO: Flexibler machen
		cmdln +=  path2k + " temp/"+ myself+ "/" + savedAs +".plan " + path2k +" temp/" +myself+ "/" + savedAs+ ".dl ";
				
		System.out.println(cmdln);
		dlvout = runDLVK(cmdln);		
		
		System.out.println(dlvout);
		
		for (String d : dlvout)	{
			if (d.startsWith("PLAN:"))
				plan = d;
		}

		//System.out.println("Plan: "+ plan);
		
		// Parse output of DLV-K so it can be parsed by parseModel
		if (plan != null){
			plan = plan.replaceAll(";", ",");
			plan = plan.replace("PLAN: " , "Best model: {");
			plan += "}";	
			planset = parseModel(plan);
			System.out.println("Planset: " + planset);
			plan = "";
			for (ELPLiteral l : planset){
				String tmp = l.getAtom().getPredicate();
				if (tmp.equals("single_move")||tmp.equals("big_move")){
					plan+=computeDirection(l) + " ";					
				} else {
					plan += l.toString();
				}	
			}
		}
		System.out.println("Plan "+plan);		
		
		return planset;
	}
	
	
private String computeDirection(ELPLiteral lit) {
		String ret = "";

		ELPAtom atom = lit.getAtom();
		if(atom.getPredicate().equals("single_move")){
			if (Integer.valueOf(atom.getTerm(2)) == Integer.valueOf(atom.getTerm(4))+1){
				ret += "north";
			}
			if (Integer.valueOf(atom.getTerm(2)) == Integer.valueOf(atom.getTerm(4))-1){
				ret += "south";
			}
			if (Integer.valueOf(atom.getTerm(1)) == Integer.valueOf(atom.getTerm(3))+1){
				ret += "west";
			}
			if (Integer.valueOf(atom.getTerm(1)) == Integer.valueOf(atom.getTerm(3))-1){
				ret += "east";
			}
		}
		if(atom.getPredicate().equals("big_move")){
			int dir = Integer.valueOf(atom.getTerm(0));
			int xStart, yStart, xGoal, yGoal;
			xStart = Integer.valueOf(atom.getTerm(1));
			yStart = Integer.valueOf(atom.getTerm(2));
			xGoal = Integer.valueOf(atom.getTerm(3));
			yGoal = Integer.valueOf(atom.getTerm(4));
			
			System.out.println(lit + " Richtung: " + dir + " Start: " +xStart +yStart  + " Ziel: " + xGoal + yGoal);
			
			switch (dir){
			case 1: 
				for (int i = yStart; i < yGoal; i++){
					ret += "north; ";
				}
				break;
			case 2: 
				for (int i = yStart; i < yGoal; i++){
					ret += "northeast; ";
				}
				break;
			case 3: 
				for (int i = xStart; i < xGoal; i++){
					ret += "east; ";
				}	
				break;
			case 4: 
				for (int i = yStart; i > yGoal; i--){
					ret += "southeast; ";
				}
				break;
			case 5: 
				for (int i = yStart; i > yGoal; i--){
					ret += "south; ";
				}
				break;
			case 6: 
				for (int i = yStart; i > yGoal; i--){
					ret += "southwest; ";
				}
				break;
			case 7: 
				for (int i = xStart; i > xGoal; i--){
					ret += "west; ";
				}
				break;
			case 8: 
				for (int i = yStart; i < yGoal; i++){
					ret += "northwest; ";
				}
				break;
			}			
		}
		System.out.println(ret);
			
		return ret;
	}

	

	protected List<String> runDLVK (String cmdln) {

		//System.out.println(cmdln);
		String text = "";
		StringBuffer errors = new StringBuffer();
		List<String> ret = new LinkedList<String>();
		
		try {
			// setup call to dlv
			Process process = Runtime.getRuntime().exec(cmdln);
			
			BufferedReader inputStream = new BufferedReader(
					new InputStreamReader(process.getInputStream()));
			BufferedReader errorStream = new BufferedReader(
					new InputStreamReader(process.getErrorStream()));
			
			// store result
			while ((text = inputStream.readLine()) != null) {
				ret.add(text);
			}

			// check for errors
			text = "";
			while ((text = errorStream.readLine()) != null) {
				errors.append(text);
			}

			//System.out.println("3");
			inputStream.close();
			errorStream.close();
			
			process = null;
		} catch (IOException e) {
			System.err.println("error!!");
			System.err.println(e);
			e.printStackTrace();
		}
		if(!"".equals(errors.toString())){
			System.err.println("error!");
			System.err.println(errors);
		}
		//System.out.println("e: "+ret);
		return ret;
	}

}

